@@ -1,4 +1,5 @@
 # -*- coding: utf-8 -*-
+import time
 import functools
 import itertools
 import random
@@ -24,20 +25,15 @@ import io
 import abc
 import collections
 
-try:
-    from dhttp import tcp
-
-except ImportError:
-    import tcp
-
-try: 
-    from dhttp.dhttpversion import DHTTP_VERSION
-
-except ImportError:
-    from dhttpversion import DHTTP_VERSION
+from . import tcp
+from .dhttpversion import DHTTP_VERSION
+from queue import Queue
+from typing import Union
 
 tcpserver = tcp
 
+NO_DEFAULT = type('DefaultSentinel', (), {})()
+
 
 # ==== BASIC WEBSOCKET INFRASTRUCTURE ====
 
@@ -104,6 +100,33 @@ for opc in WSOpcode:
 
 # --- Byte Stream Wrapping Utility for WebSockets (BSWU-WS)
 
+class Cookie(object):
+    def __init__(self, name, value, **kwargs):
+        self.name = name
+        self.value = value
+        self.options = {k.upper(): v for k, v in kwargs}
+        self.when = time.time()
+        
+    def __getitem__(self, k):
+        return self.options[k.upper()]
+
+    def __setitem__(self, k, v):
+        self.options[k.upper()] = v
+
+    def __contains__(self, k):
+        return k.upper() in self.options
+
+    def valid(self):
+        if 'Expires' in self:
+            if time.time() > time.strptime(self['Expires'], "%a, %d %b %Y %H:%M:%S %Z"):
+                return False
+
+        if 'Max-Age' in self:
+            if time.time() > self.when + float(self['Max-Age']):
+                return False
+
+        return True
+
 class WSByteStreamWrapperMixin(object):
     def _init_wrapped(self, base_class, mask = True, server = None):
         self.base_class = base_class
@@ -286,7 +309,7 @@ class WSPacketWriteStream(WSBasePacketStream):
         return True
 
     def close(self, rsv):
-        self.write_packets(WSPacket(rsv = rsv))
+        self.write_packets(WSPacket(opcode = WSOpcode.CLOSE, rsv = rsv))
         super().close()
 
     def write_packets(self, *packets):
@@ -334,7 +357,7 @@ class WSPacket(object):
         return len(self.content)
 
     def __repr__(self):
-        return f"WSPacket(fin={self.fin} rsv={bin(self.rsv).zfill(3)} mask={self.mask.hex()} opcode={self.opcode.name} len={len(self)})"
+        return "WSPacket(fin={fin} rsv={rsv} mask={mask} opcode={opcode} len={len})".format(fin=self.fin, rsv=bin(self.rsv).zfill(3), mask=(self.mask.hex() if self.mask is not None else None), opcode=self.opcode.name, len=len(self))
 
     def encode(self):
         opcode = self.opcode
@@ -356,7 +379,7 @@ class WSPacket(object):
             length_field = b''
             length_bits = len(self) & 0x7F
 
-        data += struct.pack('B', ((self.mask != None) << 7) | length_bits) + length_field
+        data += struct.pack('B', (int(self.mask is not None) << 7) | length_bits) + length_field
 
         if self.mask is not None:
             data += bytearray(self.mask) + bytearray(d ^ self.mask[i % 4] for i, d in enumerate(self.content))
@@ -364,7 +387,7 @@ class WSPacket(object):
         else:
             data += self.content
 
-        return data
+        return bytes(data)
 
     @classmethod
     def decode(self, data):
@@ -427,7 +450,12 @@ class WSPacket(object):
                 content = bytearray(b ^ mask[i % 4] for i, b in enumerate(content))
 
             if opcode == WSOpcode.TEXT:
-                content = content.decode('utf-8')
+                try:
+                    content = content.decode('utf-8')
+
+                except UnicodeDecodeError:
+                    print(content)
+                    raise
 
             pkt = WSPacket(fin, opcode, content, mask, rsv)
             messages.append(pkt)
@@ -460,7 +488,11 @@ class DHTTPLog(abc.ABC):
         pass
 
     def __str__(self):
-        return f'[{self.when.strftime("%Y-%m-%d %H:%M:%S")}]   {("(" + self.log_type() + ")").ljust(15)} {self.log_data()}'
+        return "[{time}]  {msg} {body}".format(
+            time=self.when.strftime("%Y-%m-%d %H:%M:%S"),
+            msg=("(" + self.log_type() + ")").ljust(15),
+            body=self.log_data()
+        )
 
     def __iter__(self):
         return iter((self.log_type(), self.log_data()))
@@ -493,7 +525,7 @@ class DHTTPRequestLog(DHTTPLog):
         return 'HTTP: ' + self.method
 
     def log_data(self):
-        return f'{self.ip} @ {self.request_path}'
+        return "{ip} @ {path}".format(ip=self.ip, path=self.request_path)
 
 
 class DHTTPRequest(object):    
@@ -521,6 +553,14 @@ class DHTTPRequest(object):
         self.query = query
         self.query_dict = {a: b for i, (a, b) in enumerate(query) if [x[0] for x in query].index(a) == i}
 
+    def get_cookie(self, cookie_name):
+        for name, val in self.headers:
+            if name.upper() == 'COOKIE':
+                cookies = dict(cookie.split('=') for cookie in re.split(r';\s*', val))
+
+                if cookie_name in cookies:
+                    return cookies[cookie_name]
+
     def resolve_path(self):
         path = self.path
 
@@ -543,7 +583,11 @@ class DHTTPRequest(object):
         Returns:
             str -- The header's content.
         """
-        return self.headers.get(name.upper(), '')
+        try:
+            return next(v for x, v in self.headers if x == name.upper())
+
+        except StopIteration:
+            return ''
 
 class DHTTPResponse(object):
     """
@@ -553,9 +597,20 @@ class DHTTPResponse(object):
     """
     
 
+    def set_cookie(self, name, value, expire: datetime.datetime = None, max_age: Union[int, float] = None):
+        params = []
+
+        if expire:
+            params.append('; Expires=' + expire.strftime("%a, %d %b %Y %H:%M:%S %Z"))
+
+        if max_age:
+            params.append('; Max-Age=' + str(max_age))
+
+        self.set('Set-Cookie', '{}={}{}'.format(name, value, ''.join(params)))
+
     def __init__(self, app, request, headers, write):
         self.app = app
-        self.headers = headers
+        self.headers = list(headers)
         self.request = request
         self.__write = write
         self.content = tcpserver.DequeBytesIO()
@@ -607,7 +662,11 @@ class DHTTPResponse(object):
             name {str} -- The name of the header.
             value {str} -- The value of the header.
         """
-        self.headers[name] = value
+        self.headers.append((name, value))
+
+    def setdefault(self, name, value):
+        if name not in (x for x, _ in self.headers):
+            self.set(name, value)
 
     def write(self, string):
         """
@@ -656,7 +715,7 @@ class DHTTPResponse(object):
                 call_str += ', '.join(type(x).__name__ for x in args)
 
                 if len(kwargs) > 0:
-                    call_str += ', ' + ', '.join(f'{x[0]} = {type(x[1]).__name__}' for x in kwargs.items())
+                    call_str += ', ' + ', '.join('{key} = {val}'.format(key, type(val).__name__) for key, val in kwargs.items())
 
                 call_str += ')'
 
@@ -696,21 +755,21 @@ class DHTTPResponse(object):
         self.close()
 
         size_length = len(self.content)
-        self.headers['Content-Length'] = size_length
+        self.set('Content-Length', size_length)
 
         # finish up and send response headers
         if self._status is None:
             self._write(b'HTTP/1.1 200 OK\r\n')
 
         else:
-            self._write(f'HTTP/1.1 {self._status}\r\n'.encode('utf-8'))
+            self._write('HTTP/1.1 {status}\r\n'.format(status=self._status).encode('utf-8'))
 
         h = hashlib.md5()
         h.update(self.content.read_all())
-        self.headers.setdefault('Content-MD5', base64.b64encode(h.digest()).decode('utf-8'))
-        self.headers.setdefault('Content-Type', 'text/html; encoding=utf-8')
+        self.setdefault('Content-MD5', base64.b64encode(h.digest()).decode('utf-8'))
+        self.setdefault('Content-Type', 'text/html; encoding=utf-8')
 
-        for name, value in self.headers.items():
+        for name, value in self.headers:
             self._write('{name}: {value}\r\n'.format(name = name, value = str(value).strip('\r\n')).encode('utf-8'))
 
         # send response content
@@ -723,13 +782,24 @@ class HTTPHandler(object):
 
         self.dhttp_reset()
 
+    def get_header(self, name, default = NO_DEFAULT):
+        try:
+            return next(v for x, v in self.headers if x.upper() == name.upper())
+
+        except StopIteration:
+            if default is not NO_DEFAULT:
+                return default
+
+            else:
+                raise
+
     def dhttp_reset(self, write = None, status_data = None):
         if status_data is not None:
             write(status_data)
 
         self.method = None
         self.path = None
-        self.headers = {}
+        self.headers = []
         self.content = tcpserver.DequeBytesIO()
         self.state = "BEGIN"
         self._line = b''
@@ -805,7 +875,7 @@ class HTTPHandler(object):
 
         elif self.state == 'HEADERS':
             try:
-                header_line = data.decode('utf-8')
+                header_line = data.decode('ascii')
 
             except UnicodeDecodeError:
                 self.dhttp_reset(writer, b'HTTP/1.1 400 BAD REQUEST\r\n\r\n')
@@ -815,12 +885,12 @@ class HTTPHandler(object):
             if len(header_line) == 0:
                 # F: check if the 'Content-Length' header is not null
                 try:
-                    if int(self.headers.get('CONTENT-LENGTH', 0)) <= 0:
+                    if int(self.get_header('CONTENT-LENGTH', 0)) <= 0:
                         await self.handle_http_request(writer, client)
                         self.state = 'BEGIN'
 
                     else:
-                        self._content_length = int(self.headers['CONTENT-LENGTH'])
+                        self._content_length = int(self.get_header('CONTENT-LENGTH'))
                         self.line_buffering = False
                         self.state = 'CONTENT'
 
@@ -840,12 +910,12 @@ class HTTPHandler(object):
 
                 else:
                     (name, value) = match.groups()
-                    self.headers[name.upper()] = value
+                    self.headers.append((name.upper(), value))
 
         elif self.state == 'CONTENT':
             self.content.write(data)
 
-            if len(self.content) >= self._content_length:
+            if len(self.content) >= self._content_length or client:
                 await self.handle_http_request(writer, client)
                 self.line_buffering = True
 
@@ -1035,13 +1105,13 @@ class DHTTPServer(object):
 
             path = req.resolve_path()
                 
-            headers = {}
+            headers = []
 
             if req.method in self.paths:
                 for (cbpath, request_handler) in self.paths[req.method]:
                     if fnmatch.fnmatch(path, cbpath):
                         for header in request_handler.headers:
-                            headers[header[0]] = headers[1]
+                            headers.append(tuple(header,))
 
             res = DHTTPResponse(self, req, headers, write)
 
@@ -1063,7 +1133,7 @@ class DHTTPServer(object):
                         break
 
                     else:
-                        print(DHTTPGenericLog('WARN', f'Unknown middleware return flag: {repr(flag)}\n    Expected either of: nothing/None (to continue), STOP, BREAK.\n    Ignoring and continuing...'))
+                        print(DHTTPGenericLog('WARN', 'Unknown middleware return flag: {flag}\n    Expected either of: nothing/None (to continue), STOP, BREAK.\n    Ignoring and continuing...'.format(repr(flag))))
                         continue
 
             found = False
@@ -1096,7 +1166,10 @@ class DHTTPServer(object):
 
             if not found or not res.closed:
                 res.status('404 NOT FOUND')
-                res.end()
+                res.end('The path "{path}" does not match any endpoint of this web place\'s underlying dhttp v{ver} server.'.format(
+                    path=path,
+                    ver=DHTTP_VERSION
+                ))
 
         except (BrokenPipeError, ConnectionResetError):
             return
@@ -1170,11 +1243,11 @@ class DHTTPServer(object):
         if isinstance(headers, dict):
             headers = headers.items()
 
-        def __decorator__(callback):
+        def _decorator(callback):
             self.paths.setdefault(method, []).append((path, DHTTPCallback(headers, callback)))
             return callback
 
-        return __decorator__
+        return _decorator
 
     def use(self, path = None):
         """
@@ -1214,11 +1287,11 @@ class DHTTPServer(object):
             path {str} -- An fnmatch mask of all HTTP paths to preprocess.
                           (default: {None} -- Preprocess everything!)
         """
-        def __decorator__(callback):
+        def _decorator(callback):
             self.middleware.append((path, callback))
             return callback
 
-        return __decorator__
+        return _decorator
 
     def unuse(self, func):
         """
@@ -1267,11 +1340,11 @@ class DHTTPServer(object):
             Function --   What Python sees as the actual decorator.
         """
 
-        def __decorator__(callback):
+        def _decorator(callback):
             self.on('GET', path, headers)(callback)
             return callback
 
-        return __decorator__
+        return _decorator
 
     def post(self, path, headers = ()):
         """
@@ -1303,11 +1376,11 @@ class DHTTPServer(object):
             Function --   What Python sees as the actual decorator.
         """
 
-        def __decorator__(callback):
+        def _decorator(callback):
             self.on('POST', path, headers)(callback)
             return callback
 
-        return __decorator__
+        return _decorator
 
     
     def websocket(self, path, allowed_hosts = None):
@@ -1352,7 +1425,7 @@ class DHTTPServer(object):
                         nice_host = True
 
                     elif (
-                        req.get_header('Host') in (x + str(self.port) for x in ('127.0.0.1:{port}', 'localhost:{port}')) and
+                        req.get_header('Host') in (x + str(p) for x in ('127.0.0.1:{port}', 'localhost:{port}') for (p, _, _, _) in self.tcp_servers) and
                         req.client.address[0] in ('localhost', '127.0.0.1')
                     ):
                         nice_host = True
@@ -1434,11 +1507,11 @@ class DHTTPServer(object):
             Function --   What Python sees as the actual decorator.
         """
 
-        def __decorator__(callback):
+        def _decorator(callback):
             self.on('PUT', path, headers)(callback)
             return callback
 
-        return __decorator__
+        return _decorator
 
     def serve(self, path, filename, headers = ()):
         """
@@ -1446,7 +1519,6 @@ class DHTTPServer(object):
         handler, e.g. for download.
 
             http__boring_document = app.serve('/serious_business.pdf', 'I_hope_my_boss_doesn't_see_this_filename.pdf')
-        
         Arguments:
             path {str} -- The HTTP path at the which to serve this file.
             filename {[type]} -- The path to the file to serve.
@@ -1466,14 +1538,19 @@ class DHTTPServer(object):
         """
 
         @self.get(path, headers)
-        def __callback__(req: DHTTPRequest, res: DHTTPResponse):
+        def _callback(req: DHTTPRequest, res: DHTTPResponse):
             if os.path.isfile(filename):
                 res.set('Content-Type', mimetypes.guess_type(filename)[0])
 
                 with open(filename, 'rb') as fp:
                     res.end(fp.read())
 
-        return __callback__
+        @self.on('HEAD', path, headers)
+        def _header_callback(req: DHTTPRequest, res: DHTTPRequest):
+            res.set('Content-Type', mimetypes.guess_type(filename)[0])
+            res.end()
+
+        return _callback
 
     def jinja(self, path, template, headers = (), method = 'GET'):
         _temp = template
@@ -1525,7 +1602,7 @@ class DHTTPServer(object):
                 The decorator. Decorated functions become a
             DHTTP handler, which can be then remove()'d.
         """
-        def __decorator__(param_callback):
+        def _decorator(param_callback):
             """
             The actual decorator returned by the function.
             
@@ -1558,7 +1635,7 @@ class DHTTPServer(object):
                 template = jinja2.Template(template)
 
             @self.on(method, path, headers)
-            async def __callback__(req, res):
+            async def _callback(req, res):
                 params = {}
 
                 if is_callback:
@@ -1581,9 +1658,9 @@ class DHTTPServer(object):
                 else:
                     res.end(template.render(**dict(param_callback)))
 
-            return __callback__
+            return _callback
 
-        return __decorator__
+        return _decorator
 
     def jinja_folder(self, path, folder_path, headers = (), method = 'GET'):
         """
@@ -1635,7 +1712,7 @@ class DHTTPServer(object):
                 The decorator. Decorated functions become a
             DHTTP handler, which can be then remove()'d.
         """
-        def __decorator__(param_callback):
+        def _decorator(param_callback):
             """
             The actual decorator returned by the function.
             
@@ -1666,9 +1743,17 @@ class DHTTPServer(object):
                 loader = jinja2.FileSystemLoader(os.path.join('.', folder_path)),
                 autoescape = jinja2.select_autoescape(['html', 'xml'])
             )
+            
+            @self.on('HEAD', os.path.join(path, '*'), headers)
+            async def _callback(req, res):
+                fpath = os.path.join(folder_path, os.path.relpath(req.path, path))
+
+                if os.path.isfile(fpath):
+                    res.set('Content-Type', mimetypes.guess_type(fpath)[0])
+                    res.end()
 
             @self.on(method, os.path.join(path, '*'), headers)
-            async def __callback__(req, res):
+            async def _callback(req, res):
                 fpath = os.path.join(folder_path, os.path.relpath(req.path, path))
 
                 if os.path.isfile(fpath):
@@ -1705,9 +1790,9 @@ class DHTTPServer(object):
                     else:
                         res.end(template.render(**dict(param_callback)))
 
-            return __callback__
+            return _callback
 
-        return __decorator__
+        return _decorator
 
     def static(self, path, folder_path, headers = ()):
         """
@@ -1736,7 +1821,7 @@ class DHTTPServer(object):
         """
 
         @self.get(os.path.join(path, '*'), headers)
-        def __callback__(req, res):
+        def _callback(req, res):
             fpath = os.path.join(folder_path, os.path.relpath(req.path, path))
 
             if os.path.isfile(fpath):
@@ -1745,7 +1830,7 @@ class DHTTPServer(object):
                 with open(fpath, 'rb') as fp:
                     res.end(fp.read())
 
-        return __callback__
+        return _callback
 
     def alias(self, path, new_path):
         """
@@ -1767,8 +1852,19 @@ class DHTTPClientResponse(object):
         self.headers = headers
         self.content = content
 
-    def has(self,header_name):
-        return header_name.upper() in self.headers
+    def get_cookie(self, cookie_name):
+        for name, val in self.headers:
+            if name.upper() == 'COOKIE':
+                cookies = dict(cookie.split('=') for cookie in re.split(r';\s*', val))
+
+                if cookie_name in cookies:
+                    return cookies[cookie_name]
+
+    def has(self, header_name):
+        return header_name.upper() in (x for x, _ in self.headers)
+
+    def __hasitem__(self, hn):
+        return self.has(hn)
 
     def get(self, header_name):
         """
@@ -1781,7 +1877,7 @@ class DHTTPClientResponse(object):
             str -- Data in the header.
         """
 
-        return self.headers.get(header_name.upper(), '')
+        return [v for x, v in self.headers if x == header_name.upper()]
 
     def status_number(self):
         """
@@ -1804,7 +1900,7 @@ class DHTTPClientResponseReader(object):
 
     def reset(self):
         self.status = None
-        self.headers = {}
+        self.headers = []
         self.content = tcp.DequeBytesIO()
 
         self.state = "BEGIN"
@@ -1854,6 +1950,17 @@ class DHTTPClientResponseReader(object):
         except (BaseException, UnicodeError) as err:
             traceback.print_exc()
 
+    def get_header(self, name, default = NO_DEFAULT):
+        try:
+            return next(v for x, v in self.headers if x.upper() == name.upper())
+
+        except StopIteration:
+            if default is not NO_DEFAULT:
+                return default
+
+            else:
+                raise
+
     async def handle(self, data):
         if self.state == 'BEGIN':
             status_line = re.match(r'^HTTP/([^ ]+) (.+)$', data)
@@ -1862,31 +1969,31 @@ class DHTTPClientResponseReader(object):
                 (ver, status) = status_line.groups()
 
                 if ver != '1.1':
-                    warnings.warn(ValueError(f'Unsupported HTTP version: {repr(ver)}'))
+                    warnings.warn(ValueError('Unsupported HTTP version: {ver}'.format(ver=repr(ver))))
 
                 else:
                     self.status = status
                     self.state = 'HEADERS'
 
             else:
-                warnings.warn(ValueError(f'Invalid HTTP response status line: {repr(data)}'))
+                warnings.warn(ValueError('Invalid HTTP response status line: {data}'.format(data=repr(data))))
                 self.reset()
 
         elif self.state == 'HEADERS':
             if len(data) == 0:
                 try:
-                    if int(self.headers.get('CONTENT-LENGTH', 0)) <= 0:
+                    if int(self.get_header('CONTENT-LENGTH', 0)) <= 0:
                         await self.got_response()
                         self.reset()
 
                     else:
-                        self._content_length = int(self.headers['CONTENT-LENGTH'])
+                        self._content_length = int(self.get_header('CONTENT-LENGTH'))
                         self.line_buffering = False
                         self.state = 'CONTENT'
 
                 except ValueError:
                     if 'CONTENT-LENGTH' in self.headers:
-                        warnings.warn(ValueError(f'Invalid HTTP Content-Length value: {repr(self.headers["CONTENT-LENGTH"])}'))
+                        warnings.warn(ValueError('Invalid HTTP Content-Length value: {content_length}'.format(content_length=repr(self.get_header("CONTENT-LENGTH")))))
                         self.reset()
                         return
 
@@ -1898,12 +2005,12 @@ class DHTTPClientResponseReader(object):
 
                 # G: check if header matches format
                 if match is None:
-                    warnings.warn(ValueError(f"Invalid HTTP header line: {repr(data)} - ignoring."))
+                    warnings.warn(ValueError("Invalid HTTP header line: {data} - ignoring.".format(data=repr(data))))
                     return
 
                 else:
                     (name, value) = match.groups()
-                    self.headers[name.upper()] = value
+                    self.headers.append((name.upper(), value))
             
         elif self.state == 'CONTENT':
             self.content.write(data)
@@ -1921,7 +2028,7 @@ class DHTTPClientResponseReader(object):
                 return
 
     async def got_response(self):
-        resp = DHTTPClientResponse(self.status, self.headers, self.content)
+        resp = DHTTPClientResponse(self.status, self.headers, self.content.read_all())
 
         async def sync_handler(f):
             f(resp)
@@ -1947,18 +2054,21 @@ class DHTTPClient(object):
         port {int} -- The port of the server to connect to. (default: {80})
     """
 
-    def __init__(self, host, port = 80):
+    def __init__(self, host: str, port: int = 80):
         self.host = host
         self.port = port
 
         self.client = tcp.AsyncTCPClient()
         self.running = False
+        self.cookies = {}
         self.connected = False
         self.response_reader = DHTTPClientResponseReader(self.client)
 
         self._on_response = set()
         self._resp = None
 
+        self.sequential_handlers = Queue()
+
         self.response_reader.on_response(self.got_response)
 
     def reset_response_handlers(self):
@@ -1987,7 +2097,7 @@ class DHTTPClient(object):
             self.running = True
 
             if self.client.connected:
-                await self.client.run()
+                await self.client.loop()
 
             else:
                 await self.client.run_connect(self.host, self.port, *args, **kwargs)
@@ -1997,7 +2107,7 @@ class DHTTPClient(object):
             self.running = False
             self.client.stop()
 
-    async def request(self, path = '/', method = 'GET', headers = (), data = b'', tls = False, cert_mode = tcpserver.CertMode.CM_GENERATED, fn_component = '', cert_file = None, key_file = None, key_size = 2048):
+    async def request(self, path = '/', method = 'GET', headers = (), data = b'', tls = False, cert_mode = tcpserver.CertMode.CM_GENERATED, fn_component = '', cert_file = None, key_file = None, key_size = 2048, handler = None, auto_start = False):
         if fn_component:
             if key_file is None:
                 key_file = ".{}.key.pem".format(fn_component)
@@ -2008,10 +2118,10 @@ class DHTTPClient(object):
         if not self.client.connected:
             await self.client.connect(self.host, self.port, tls = tls, cert_mode = cert_mode, cert_file = cert_file, key_file = key_file, key_size = key_size)
 
-        headers = dict(headers)
+        headers = list(headers)
         
-        host = f"{self.host}{'' if self.port == 80 else ':' + str(self.port)}"
-        origin = f"http://{self.host}:{self.port}"
+        host = "{host}{port}".format(host=self.host, port=('' if self.port == 80 else ':' + str(self.port)))
+        origin = "http://{host}:{port}".format(host=self.host, port=self.port)
 
         if isinstance(data, str):
             data = data.encode('utf-8')
@@ -2024,27 +2134,63 @@ class DHTTPClient(object):
 
         h = hashlib.md5(data)
 
+        cookies = []
+
+        for cn, cv in self.cookies.items():
+            if cv.valid():
+                cookies.append('{}={}'.format(cn, cv.value))
+
+        if cookies:
+            headers.append(('Cookie', '; '.join(cookies)))
+
         if data:
-            headers.setdefault('Content-Length', len(data))
-            headers.setdefault('Content-MD5', base64.b64encode(h.digest()).decode('utf-8'))
+            headers.append(('Content-Length', len(data)))
+            headers.append(('Content-MD5', base64.b64encode(h.digest()).decode('utf-8')))
 
-        headers.setdefault('Host', host)
-        headers.setdefault('Origin', origin)
+        headers.append(('Host', host))
+        headers.append(('Origin', origin))
 
-        self.client.write(f'{method} {path.strip(" ")} HTTP/1.1\r\n'.encode('utf-8'))
+        self.client.write('{method} {path} HTTP/1.1\r\n'.format(method=method, path=path.strip(' ')).encode('utf-8'))
 
-        for name, value in headers.items():
+        for name, value in headers:
             self.client.write('{name}: {value}\r\n'.format(name = name, value = str(value).strip('\r\n')).encode('utf-8'))
 
-        self.client.write(b'\r\n' + data)
+        if handler:
+            self.sequential_handlers.put_nowait(handler)
+
+        else:
+            self.sequential_handlers.put_nowait(None)
+
+        self.client.write(b'\r\n' + data + b'\r\n')
+
+        if auto_start:
+            await self.start()
+            return self._resp
+
+    async def got_response(self, resp: DHTTPClientResponse):
+        if resp.has('Set-Cookie'):
+            for c in resp.get('Set-Cookie'):
+                n = c[:c.index('=')]
+                v = c[c.index('=') + 1:]
+
+                data = v.split(';')
+                self.cookies[n] = Cookie(n, data[0], **{k: v for k, v in (re.split(r';\s*', d) for d in data[1:])})
 
-    async def got_response(self, resp):
         self._resp = resp
 
         async def sync_handler(f):
             f(resp)
 
         async with trio.open_nursery() as nursery:
+            handler = self.sequential_handlers.get_nowait()
+
+            if handler:
+                if inspect.iscoroutinefunction(handler):
+                    nursery.start_soon(handler, resp)
+
+                else:
+                    nursery.start_soon(sync_handler, handler)
+                
             for f in self._on_response:
                 if inspect.iscoroutinefunction(f):
                     nursery.start_soon(f, resp)
@@ -2052,6 +2198,8 @@ class DHTTPClient(object):
                 else:
                     nursery.start_soon(sync_handler, f)
 
+        self.stop()
+
 
 class BaseDHTTPException(BaseException):
     pass
@@ -2129,7 +2277,7 @@ class DHTTPWebsocketClient(object):
         if resp.status_number() != 101:
             raise HTTPStatusError('Got a bad or invalid HTTP status while trying to upgrade to a WebSocket connection:  ' + resp.status)
 
-        elif not resp.has('Upgrade') or resp.get('Upgrade').upper() != 'WEBSOCKET':
+        elif not resp.has('Upgrade') or resp.get('Upgrade')[0].upper() != 'WEBSOCKET':
             raise HTTPStatusError('Got no HTTP upgrade, or an invalid one, instead of a WebSocket upgrade:  ' + ('(none)' if not resp.has('Upgrade') else repr(resp.get('Upgrade'))))
 
         elif not resp.has('Sec-WebSocket-Accept'):
@@ -2137,19 +2285,22 @@ class DHTTPWebsocketClient(object):
 
         else:
             try:
-                acc = base64.b64decode(resp.get('Sec-WebSocket-Accept'))
+                acc = base64.b64decode(resp.get('Sec-WebSocket-Accept')[0])
 
             except binascii.Error:
-                raise WSHandshakeError(f'Sec-WebSocket-Accept header in the response is not a valid Base64 value!')
+                raise WSHandshakeError('Sec-WebSocket-Accept header in the response is not a valid Base64 value!')
 
             else:
                 if acc != self.ws_accept:
-                    raise WSHandshakeError(f'The response\'s Sec-WebSocket-Accept header value does not match the predicted key (expected {base64.b64encode(self.ws_accept)}, got {base64.b64encode(acc)})!')
+                    raise WSHandshakeError('The response\'s Sec-WebSocket-Accept header value does not match the predicted key (expected {expec}, got {real})!'.format(
+                        expec=base64.b64encode(self.ws_accept).decode('utf-8'),
+                        real=base64.b64encode(acc.decode('utf-8'))
+                    ))
 
                 elif resp.has('Sec-WebSocket-Extensions'):
                     raise WSHandshakeError('Sec-WebSocket-Extensions is not supported!')
 
-                elif resp.has('Sec-WebSocket-Protocol') != bool(self.ws_subprotocol) or (self.ws_subprotocol is not None and resp.get('Sec-Websocket-Protocol') != self.ws_subprotocol):
+                elif resp.has('Sec-WebSocket-Protocol') != bool(self.ws_subprotocol) or (self.ws_subprotocol is not None and resp.get('Sec-Websocket-Protocol')[0] != self.ws_subprotocol):
                     raise WSHandshakeError('The response\'s value of Sec-WebSocket-Protocol does not match the requested value!')
 
                 else:
@@ -2238,14 +2389,14 @@ class DHTTPWebsocketClient(object):
 
         self.ws_subprotocol = subprotocol
 
-        headers = {
+        headers = list({
             'Connection': 'keep-alive, Upgrade',
             'Upgrade': 'websocket',
             'Sec-WebSocket-Key': ws_key_header,
             'Sec-WebSocket-Version': 13
-        }
+        }.items())
 
         if subprotocol is not None:
-            headers['Sec-WebSocket-Protocol'] = subprotocol
+            headers.append(('Sec-WebSocket-Protocol', subprotocol))
 
         await self.conn.request(path, headers = headers, tls = tls, cert_mode = cert_mode, cert_file = cert_file, key_file = key_file, key_size = key_size)
\ No newline at end of file
