@@ -534,14 +534,17 @@ class AsyncTCPClient(object):
                     sleep_amount = 1 / 20
 
                 except ssl.SSLError as err:
-                    print(dhttp.DHTTPGenericLog('ERR', f'(SSL CLIENT: {type(err).__name__}) {str(err)}'))
+                    print(dhttp.DHTTPGenericLog('ERR', '(SSL CLIENT: {ename}) {estr}'.format(
+                        ename=type(err).__name__,
+                        estr=str(err)
+                    )))
                     break
             
             if not self.running:
-                read_amount = min(1024, unsent)
+                read_amount = min(4096, unsent)
 
             else:
-                read_amount = 1024
+                read_amount = 4096
 
             write_data = self._write_buffer.read(read_amount, seek = False)
 
@@ -550,7 +553,11 @@ class AsyncTCPClient(object):
                 self._write_buffer.truncate(sent)
                 unsent = max(0, unsent - sent) + max(0, len(write_data) - sent)
 
-            await trio.sleep(sleep_amount)
+            if unsent == 0:
+                await trio.sleep(sleep_amount)
+
+            else:
+                await trio.sleep(0)
 
         # self.socket.shutdown(socketlib.SHUT_RDWR)
         self.socket.close()
@@ -877,7 +884,10 @@ class AsyncTCPServer(object):
                 sleep_amount = 1 / 15
 
             except ssl.SSLError as err:
-                print(dhttp.DHTTPGenericLog('ERR', f'(SSL SERVER: {type(err).__name__}) {str(err)}'))
+                print(dhttp.DHTTPGenericLog('ERR', '(SSL SERVER: {ename}) {estr}'.format(
+                        ename=type(err).__name__,
+                        estr=str(err)
+                    )))
                 client.close()
                 break
             
@@ -899,7 +909,11 @@ class AsyncTCPServer(object):
                 if client._pending_close:
                     await client._close()
 
-            await trio.sleep(sleep_amount)
+            if unsent == 0:
+                await trio.sleep(sleep_amount)
+
+            else:
+                await trio.sleep(0)
 
         if not client.closed:
             await client._close()
@@ -986,7 +1000,10 @@ class AsyncTCPServer(object):
                     sleep_amount = 0.25
 
                 except ssl.SSLError as err:
-                    print(dhttp.DHTTPGenericLog('ERR', f'(SSL LISTEN: {type(err).__name__}) {str(err)}'))
+                    print(dhttp.DHTTPGenericLog('ERR', '(SSL LISTEN: {ename}) {estr}'.format(
+                        ename=type(err).__name__,
+                        estr=str(err)
+                    )))
                     sleep_amount = 0.1
                     break
 
